import {
  assertUnreachable,
  FileOrFolderPath,
  FilePath,
  FolderPath,
  FolderPathUtils,
  maybe,
  PathPatternDto
} from "@upload-io/upload-shared-iso";
import { FolderPermissionsWithPathAndScope } from "@upload-io/upload-shared-iso/folders/model/FolderPermissionsWithPathAndScope";
import { descend, sortWith } from "ramda";

export class PathPermissionUtils {
  static getEffectivePermissionsAtPattern<T>(
    pattern: PathPatternDto,
    permissionsUnsorted: Array<FolderPermissionsWithPathAndScope<T>>
  ): T | undefined {
    switch (pattern.scope) {
      case "This": {
        return PathPermissionUtils.getEffectivePermissions(pattern.path, permissionsUnsorted);
      }
      case "Children": {
        return PathPermissionUtils.getEffectivePermissions(FileOrFolderPath(`${pattern.path}/`), permissionsUnsorted);
      }
      case "Grandchildren+": {
        return PathPermissionUtils.getEffectivePermissions(FileOrFolderPath(`${pattern.path}//`), permissionsUnsorted);
      }
      default:
        assertUnreachable(pattern.scope);
    }
  }

  /**
   * From the provided permissions, returns the one that applies to the 'pathToAccess', if any.
   */
  static getEffectivePermissions<T>(
    pathToAccess: FilePath | FolderPath | FileOrFolderPath,
    permissionsUnsorted: Array<FolderPermissionsWithPathAndScope<T>>
  ): T | undefined {
    const pathToAccessAncestors = FolderPathUtils.getAncestors(pathToAccess);

    // In JavaScript, if you pass a negative to an array, you get 'undefined', which is what we want...
    const pathToAccessParent: FileOrFolderPath | undefined = maybe(
      FileOrFolderPath,
      pathToAccessAncestors[pathToAccessAncestors.length - 1]
    );
    const pathToAccessGrandParentAndUp: FileOrFolderPath[] = pathToAccessAncestors
      .slice(0, pathToAccessAncestors.length - 1)
      .map(FileOrFolderPath);

    const nearestFolderPermissions = PathPermissionUtils.getNearestPathPermissions(
      pathToAccess,
      pathToAccessAncestors,
      permissionsUnsorted
    );

    return nearestFolderPermissions.find(rule => {
      switch (rule.match.scope) {
        case "This":
          return rule.match.path === pathToAccess;
        case "Children":
          return rule.match.path === pathToAccessParent;
        case "Grandchildren+":
          return pathToAccessGrandParentAndUp.includes(rule.match.path);
        default:
          assertUnreachable(rule.match.scope);
      }
    })?.permissions;
  }

  /**
   * Returns the path permissions for the nearest ancestor.
   *
   * All items in the result will have the same 'path'.
   *
   * Intended to mimic 'effectiveFolderSettingsService.getNearestAncestor', so we get consistent behaviour between public
   * folder permissions and API key path permissions. See comment on this class: we always take the nearest ancestor's
   * permissions, even if they won't be effective on the path we're checking, even if a higher ancestor's permissions would.
   */
  private static getNearestPathPermissions<T>(
    pathToAccess: FilePath | FolderPath | FileOrFolderPath,
    pathToAccessAncestors: FolderPath[],
    permissionsUnsorted: Array<FolderPermissionsWithPathAndScope<T>>
  ): Array<FolderPermissionsWithPathAndScope<T>> {
    // Sort the permissions, such that descendants appear before their ascendants.
    const permissions = sortWith([descend(x => x.match.path)], permissionsUnsorted);
    const pathsToMatch = new Set([...pathToAccessAncestors, FolderPath(pathToAccess)]);
    const nearestFolderPath = permissions.find(x => pathsToMatch.has(FolderPath(x.match.path)))?.match.path;
    return permissions.filter(x => x.match.path === nearestFolderPath);
  }
}
