import {
  OrderSensitiveQueryParams,
  QueryParam
} from "@upload-io/upload-shared-iso/common/model/OrderSensitiveQueryParams";
import { JsonValue, optional } from "@upload-io/upload-shared-iso";
import { sortBy } from "ramda";
import { JSONSchema7 } from "json-schema";
import { QueryParamDefinitionFactory } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/QueryParamDefinitionFactory";
import { JsonSchemaUtils } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/JsonSchemaUtils";
import { NormalizedJsonSchema } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/ObjectWithMetadata";
import { IsComplete } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/IsComplete";
import { QueryParamEncoder } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/QueryParamEncoder";

export class JsonToQueryStringFunction {
  constructor(private readonly rootSchema: JSONSchema7) {}

  toQueryString(params: JsonValue): OrderSensitiveQueryParams {
    return this.toQueryStringInternal(params, this.rootSchema).value;
  }

  private toQueryStringInternal(value: JsonValue, schemaRaw: JSONSchema7): IsComplete<OrderSensitiveQueryParams> {
    const schemas = JsonSchemaUtils.normalizePolymorphisms(schemaRaw);
    const possibleParams = schemas.map(x => this.toQueryStringNormalized(value, x));
    const completeParams = possibleParams.filter(x => x.isComplete);
    const isComplete = completeParams.length > 0;
    return {
      isComplete,
      value: completeParams[0]?.value ?? possibleParams[0].value
    };
  }

  private toQueryStringNormalized(
    value: JsonValue,
    schema: NormalizedJsonSchema
  ): IsComplete<OrderSensitiveQueryParams> {
    const queryParamName = (): string[] => optional(QueryParamDefinitionFactory.getQueryParamName(schema));
    const handleChildResults = (
      results: Array<IsComplete<OrderSensitiveQueryParams>>
    ): IsComplete<OrderSensitiveQueryParams> => ({
      isComplete: results.every(x => x.isComplete),
      value: results.flatMap(x => x.value)
    });

    if (value === null) {
      return {
        isComplete: JsonSchemaUtils.passesBasicTypeCheck(schema, value),
        value: queryParamName().map((name): QueryParam => ({ name, value: "null" }))
      };
    }

    if (Array.isArray(value)) {
      const itemSchema = schema.items;
      if (itemSchema === undefined || typeof itemSchema === "boolean" || Array.isArray(itemSchema)) {
        throw new Error("Encountered an array in the object, but the schema was not expecting an array.");
      }
      return handleChildResults(value.map(x => this.toQueryStringInternal(x, itemSchema)));
    }

    if (typeof value === "object") {
      let hasAdditionalProperties = false;
      const properties = Object.entries(value).flatMap(
        ([propertyName, propertyValue]): Array<{
          propertySchema: JSONSchema7;
          propertyValue: JsonValue;
        }> => {
          const propertySchema = schema.properties?.[propertyName];
          if (propertySchema === undefined || typeof propertySchema === "boolean" || Array.isArray(propertySchema)) {
            if (schema.additionalProperties === false) {
              // This field wasn't expected, which means this isn't the correct type for this object (which is expected,
              // as it just means we're inside a polymorphism and we've been attempted on one side of the type union,
              // and we're acknowledging that we don't fit into it).
              hasAdditionalProperties = true;
            }
            return [];
          }

          return [
            {
              propertyValue,
              propertySchema
            }
          ];
        }
      );

      const hasAllRequiredFields = (schema.required ?? []).every(x => value[x] !== undefined);
      const isOuterComplete = !hasAdditionalProperties && hasAllRequiredFields;

      const propertiesSortedByQueryName = sortBy(
        x => QueryParamDefinitionFactory.getQueryParamName(x.propertySchema) ?? "",
        properties
      );
      const childResults = handleChildResults(
        propertiesSortedByQueryName.map(({ propertyValue, propertySchema }) =>
          this.toQueryStringInternal(propertyValue, propertySchema)
        )
      );

      return {
        isComplete: isOuterComplete && childResults.isComplete,
        value: childResults.value
      };
    }

    const rewriteTo = QueryParamDefinitionFactory.getQueryParamConstantRewrite(schema);
    const rewriteFrom = schema.enum !== undefined && schema.enum.length === 1 ? schema.enum[0] : undefined;
    const queryValuePreEncode = rewriteFrom === value && rewriteTo !== undefined ? rewriteTo : value;

    // Scalars...
    return {
      isComplete: JsonSchemaUtils.passesBasicTypeCheck(schema, value),
      value: queryParamName().map(
        (name): QueryParam => ({
          name,
          value: QueryParamEncoder.encodeForQueryString(queryValuePreEncode.toString(), schema)
        })
      )
    };
  }
}
