import {
  FileName,
  FileOrFolderPath,
  FolderName,
  FolderPath,
  maybe
} from "@upload-io/upload-shared-iso/common/model/Primitives";
import { FolderPathUtils } from "@upload-io/upload-shared-iso/folders/FolderPathUtils";

describe("FolderPathUtils", (): void => {
  test("getParentFolder", (): void => {
    const assert = (given: string, expected: string | undefined): void => {
      expect(FolderPathUtils.getParentFolder(FileOrFolderPath(given))).toStrictEqual(maybe(FolderPath, expected));
    };
    assert("/", undefined);
    assert("//", "/");
    assert("///", "//");
    assert("//foo", "//");
    assert("///foo", "///");
    assert("////foo", "////");
    assert("////foo/", "////foo");
    assert("////foo/bar", "////foo");
    assert("////foo//bar", "////foo/");
    assert("////foo//bar/", "////foo//bar");
    assert("////foo///bar/", "////foo///bar");
    assert("////foo////bar/", "////foo////bar");
    assert("//foo/", "//foo");
    assert("//foo//", "//foo/");
    assert("//foo///", "//foo//");
    assert("/foo", "/");
    assert("/foo/", "/foo");
    assert("/foo//", "/foo/");
    assert("/foo/bar", "/foo");
    assert("/foo/bar/", "/foo/bar");
    assert("/foo/bar//", "/foo/bar/");
    assert("/foo/bar///", "/foo/bar//");
    assert("/foo//", "/foo/");
    assert("/foo///", "/foo//");
    assert("/foo//bar", "/foo/");
    assert("/foo///bar", "/foo//");
  });

  test("getAncestorsAndSelf", (): void => {
    const assert = (given: string, expected: string[]): void => {
      expect(FolderPathUtils.getAncestors(FileOrFolderPath(given))).toStrictEqual(expected.map(x => FolderPath(x)));
    };

    assert("/", []);
    assert("//", ["/"]);
    assert("///", ["/", "//"]);
    assert("//foo", ["/", "//"]);
    assert("///foo", ["/", "//", "///"]);
    assert("////foo", ["/", "//", "///", "////"]);
    assert("////foo/", ["/", "//", "///", "////", "////foo"]);
    assert("////foo/bar", ["/", "//", "///", "////", "////foo"]);
    assert("////foo//bar", ["/", "//", "///", "////", "////foo", "////foo/"]);
    assert("////foo//bar/", ["/", "//", "///", "////", "////foo", "////foo/", "////foo//bar"]);
    assert("////foo///bar/", ["/", "//", "///", "////", "////foo", "////foo/", "////foo//", "////foo///bar"]);
    assert("////foo////bar/", [
      "/",
      "//",
      "///",
      "////",
      "////foo",
      "////foo/",
      "////foo//",
      "////foo///",
      "////foo////bar"
    ]);
    assert("//foo/", ["/", "//", "//foo"]);
    assert("//foo//", ["/", "//", "//foo", "//foo/"]);
    assert("//foo///", ["/", "//", "//foo", "//foo/", "//foo//"]);
    assert("/foo", ["/"]);
    assert("/foo/", ["/", "/foo"]);
    assert("/foo//", ["/", "/foo", "/foo/"]);
    assert("/foo/bar", ["/", "/foo"]);
    assert("/foo/bar/", ["/", "/foo", "/foo/bar"]);
    assert("/foo/bar//", ["/", "/foo", "/foo/bar", "/foo/bar/"]);
    assert("/foo/bar///", ["/", "/foo", "/foo/bar", "/foo/bar/", "/foo/bar//"]);
    assert("/foo//", ["/", "/foo", "/foo/"]);
    assert("/foo///", ["/", "/foo", "/foo/", "/foo//"]);
    assert("/foo//bar", ["/", "/foo", "/foo/"]);
    assert("/foo///bar", ["/", "/foo", "/foo/", "/foo//"]);
  });

  test("getFolderName", (): void => {
    const assert = (given: string, expected: string): void => {
      expect(FolderPathUtils.getFolderName(FolderPath(given))).toStrictEqual(FolderName(expected));
    };
    const empty = "";
    const foo = "foo";
    const bar = "bar";
    assert("/", empty);
    assert("//", empty);
    assert("///", empty);
    assert("//foo", foo);
    assert("///foo", foo);
    assert("////foo", foo);
    assert("////foo/", empty);
    assert("////foo/bar", bar);
    assert("////foo//bar", bar);
    assert("////foo//bar/", empty);
    assert("////foo///bar/", empty);
    assert("////foo////bar/", empty);
    assert("//foo/", empty);
    assert("//foo//", empty);
    assert("//foo///", empty);
    assert("/foo", foo);
    assert("/foo/", empty);
    assert("/foo//", empty);
    assert("/foo/bar", bar);
    assert("/foo/bar/", empty);
    assert("/foo/bar//", empty);
    assert("/foo/bar///", empty);
    assert("/foo//", empty);
    assert("/foo///", empty);
    assert("/foo//bar", bar);
    assert("/foo///bar", bar);
  });

  test("combine", (): void => {
    const assert = (given: string): void => {
      const folderName = FolderPathUtils.getFolderName(FolderPath(given));
      const folderParent = FolderPathUtils.getParentFolder(FileOrFolderPath(given));
      const combined = FolderPathUtils.combine(folderParent, folderName);
      expect(combined).toStrictEqual(FileOrFolderPath(given));
    };
    assert("/");
    assert("//");
    assert("///");
    assert("//foo");
    assert("///foo");
    assert("////foo");
    assert("////foo/");
    assert("////foo/bar");
    assert("////foo//bar");
    assert("////foo//bar/");
    assert("////foo///bar/");
    assert("////foo////bar/");
    assert("//foo/");
    assert("//foo//");
    assert("//foo///");
    assert("/foo");
    assert("/foo/");
    assert("/foo//");
    assert("/foo/bar");
    assert("/foo/bar/");
    assert("/foo/bar//");
    assert("/foo/bar///");
    assert("/foo//");
    assert("/foo///");
    assert("/foo//bar");
    assert("/foo///bar");
  });

  test("splitExtension", (): void => {
    const assert = (fileName: string, prefix: string, suffix: string): void => {
      expect(FolderPathUtils.splitExtension(FileName(fileName))).toStrictEqual([prefix, suffix]);
    };

    assert("", "", "");
    assert("hello", "hello", "");
    assert(".tar.gz", "", ".tar.gz");
    assert(".hello.tar.gz", ".hello", ".tar.gz");
    assert("hello.world", "hello.world", "");
    assert("hello.tar.gz", "hello", ".tar.gz");
    assert("hello.world.tar.gz", "hello.world", ".tar.gz");
    assert("hello.jpg", "hello", ".jpg");
    assert("hello.jpeg", "hello", ".jpeg");
    assert("hello.world.jpg", "hello.world", ".jpg");
    assert("hello.world.jpg.bak", "hello.world", ".jpg.bak");
    assert("hello.world.goodbye.jpg.bak", "hello.world.goodbye", ".jpg.bak");
  });
});
